import { Route } from '@modules/geography/core/domain/route.types';
import { Role } from './ad.types';
import { Point } from './value-objects/point.value-object';
import { PathCreatorException } from './match.errors';

export class PathCreator {
  constructor(
    private readonly roles: Role[],
    private readonly waypoints: Point[],
  ) {
    if (roles.length == 0)
      throw new PathCreatorException(
        new Error('At least a role must be defined'),
      );
    if (waypoints.length < 2)
      throw new PathCreatorException(
        new Error('At least 2 waypoints must be defined'),
      );
  }

  public getBasePaths = (): Path[] => {
    const paths: Path[] = [];
    if (
      this.roles.includes(Role.DRIVER) &&
      this.roles.includes(Role.PASSENGER)
    ) {
      if (this.waypoints.length == 2) {
        // 2 points => same route for driver and passenger
        paths.push(this._createGenericPath());
      } else {
        paths.push(this._createDriverPath(), this._createPassengerPath());
      }
    } else if (this.roles.includes(Role.DRIVER)) {
      paths.push(this._createDriverPath());
    } else if (this.roles.includes(Role.PASSENGER)) {
      paths.push(this._createPassengerPath());
    }
    return paths;
  };

  private _createGenericPath = (): Path =>
    this._createPath(this.waypoints, PathType.GENERIC);

  private _createDriverPath = (): Path =>
    this._createPath(this.waypoints, PathType.DRIVER);

  private _createPassengerPath = (): Path =>
    this._createPath(
      [this._firstWaypoint(), this._lastWaypoint()],
      PathType.PASSENGER,
    );

  private _firstWaypoint = (): Point => this.waypoints[0];

  private _lastWaypoint = (): Point =>
    this.waypoints[this.waypoints.length - 1];

  private _createPath = (waypoints: Point[], type: PathType): Path => ({
    type,
    waypoints,
  });
}

export type Path = {
  type: PathType;
  waypoints: Point[];
};

export type TypedRoute = {
  type: PathType;
  route: Route;
};

/**
 * PathType id used for route calculation, to reduce the number of routes to compute :
 * - a single route for a driver only
 * - a single route for a passenger only
 * - a single route for a driver and passenger with 2 waypoints given
 * - two routes for a driver and passenger with more than 2 waypoints given
 *    (all the waypoints as driver, only origin and destination as passenger as
 *      intermediate waypoints doesn't matter in that case)
 */
export enum PathType {
  GENERIC = 'generic',
  DRIVER = 'driver',
  PASSENGER = 'passenger',
}
