import { DateTimeTransformerPort } from '@modules/ad/core/application/ports/datetime-transformer.port';
import { DefaultParams } from '@modules/ad/core/application/ports/default-params.type';
import { RouteProviderPort } from '@modules/ad/core/application/ports/route-provider.port';
import { MatchQuery } from '@modules/ad/core/application/queries/match/match.query';
import { AlgorithmType } from '@modules/ad/core/application/types/algorithm.types';
import { Waypoint } from '@modules/ad/core/application/types/waypoint.type';
import { Frequency } from '@modules/ad/core/domain/ad.types';

const originWaypoint: Waypoint = {
  position: 0,
  lat: 48.689445,
  lon: 6.17651,
  houseNumber: '5',
  street: 'Avenue Foch',
  locality: 'Nancy',
  postalCode: '54000',
  country: 'France',
};
const destinationWaypoint: Waypoint = {
  position: 1,
  lat: 48.8566,
  lon: 2.3522,
  locality: 'Paris',
  postalCode: '75000',
  country: 'France',
};
const intermediateWaypoint: Waypoint = {
  position: 1,
  lat: 48.966912,
  lon: 4.3655,
  locality: 'Châlons-en-Champagne',
  postalCode: '51000',
  country: 'France',
};

const defaultParams: DefaultParams = {
  DEPARTURE_TIME_MARGIN: 900,
  DRIVER: false,
  SEATS_PROPOSED: 3,
  PASSENGER: true,
  SEATS_REQUESTED: 1,
  STRICT: false,
  TIMEZONE: 'Europe/Paris',
  ALGORITHM_TYPE: AlgorithmType.PASSENGER_ORIENTED,
  REMOTENESS: 15000,
  USE_PROPORTION: true,
  PROPORTION: 0.3,
  USE_AZIMUTH: true,
  AZIMUTH_MARGIN: 10,
  MAX_DETOUR_DISTANCE_RATIO: 0.3,
  MAX_DETOUR_DURATION_RATIO: 0.3,
};

const mockInputDateTimeTransformer: DateTimeTransformerPort = {
  fromDate: jest.fn().mockImplementation(() => '2023-08-27'),
  toDate: jest.fn().mockImplementation(() => '2023-08-27'),
  day: jest.fn().mockImplementation(() => 0),
  time: jest.fn().mockImplementation(() => '23:05'),
};

const mockRouteProvider: RouteProviderPort = {
  getBasic: jest
    .fn()
    .mockImplementationOnce(() => ({
      distance: 350101,
      duration: 14422,
      fwdAzimuth: 273,
      backAzimuth: 93,
      distanceAzimuth: 336544,
      points: [],
    }))
    .mockImplementationOnce(() => ({
      distance: 340102,
      duration: 13423,
      fwdAzimuth: 273,
      backAzimuth: 93,
      distanceAzimuth: 336544,
      points: [],
    }))
    .mockImplementationOnce(() => ({
      distance: 350101,
      duration: 14422,
      fwdAzimuth: 273,
      backAzimuth: 93,
      distanceAzimuth: 336544,
      points: [],
    }))
    .mockImplementationOnce(() => ({
      distance: 350101,
      duration: 14422,
      fwdAzimuth: 273,
      backAzimuth: 93,
      distanceAzimuth: 336544,
      points: [],
    }))
    .mockImplementationOnce(() => ({
      distance: 340102,
      duration: 13423,
      fwdAzimuth: 273,
      backAzimuth: 93,
      distanceAzimuth: 336544,
      points: [],
    }))
    .mockImplementationOnce(() => {
      throw new Error();
    }),
};

describe('Match Query', () => {
  it('should set default values', async () => {
    const matchQuery = new MatchQuery(
      {
        frequency: Frequency.PUNCTUAL,
        fromDate: '2023-08-28',
        toDate: '2023-08-28',
        schedule: [
          {
            time: '01:05',
          },
        ],
        waypoints: [originWaypoint, destinationWaypoint],
      },
      mockRouteProvider,
    );
    matchQuery
      .setMissingMarginDurations(defaultParams.DEPARTURE_TIME_MARGIN)
      .setMissingStrict(defaultParams.STRICT)
      .setDefaultDriverAndPassengerParameters({
        driver: defaultParams.DRIVER,
        passenger: defaultParams.PASSENGER,
        seatsProposed: defaultParams.SEATS_PROPOSED,
        seatsRequested: defaultParams.SEATS_REQUESTED,
      })
      .setDefaultAlgorithmParameters({
        algorithmType: defaultParams.ALGORITHM_TYPE,
        remoteness: defaultParams.REMOTENESS,
        useProportion: defaultParams.USE_PROPORTION,
        proportion: defaultParams.PROPORTION,
        useAzimuth: defaultParams.USE_AZIMUTH,
        azimuthMargin: defaultParams.AZIMUTH_MARGIN,
        maxDetourDistanceRatio: defaultParams.MAX_DETOUR_DISTANCE_RATIO,
        maxDetourDurationRatio: defaultParams.MAX_DETOUR_DURATION_RATIO,
      })
      .setDatesAndSchedule(mockInputDateTimeTransformer);
    expect(matchQuery.strict).toBeFalsy();
    expect(matchQuery.driver).toBeFalsy();
    expect(matchQuery.seatsProposed).toBe(3);
    expect(matchQuery.seatsRequested).toBe(1);
    expect(matchQuery.algorithmType).toBe(AlgorithmType.PASSENGER_ORIENTED);
    expect(matchQuery.remoteness).toBe(15000);
    expect(matchQuery.useProportion).toBeTruthy();
    expect(matchQuery.proportion).toBe(0.3);
    expect(matchQuery.useAzimuth).toBeTruthy();
    expect(matchQuery.azimuthMargin).toBe(10);
    expect(matchQuery.maxDetourDistanceRatio).toBe(0.3);
    expect(matchQuery.maxDetourDurationRatio).toBe(0.3);
    expect(matchQuery.fromDate).toBe('2023-08-27');
    expect(matchQuery.toDate).toBe('2023-08-27');
    expect(matchQuery.schedule[0].day).toBe(0);
    expect(matchQuery.schedule[0].time).toBe('23:05');
    expect(matchQuery.schedule[0].margin).toBe(900);
  });

  it('should set good values for seats', async () => {
    const matchQuery = new MatchQuery(
      {
        frequency: Frequency.PUNCTUAL,
        fromDate: '2023-08-28',
        toDate: '2023-08-28',
        seatsProposed: -1,
        seatsRequested: -1,
        schedule: [
          {
            time: '07:05',
          },
        ],
        waypoints: [originWaypoint, destinationWaypoint],
      },
      mockRouteProvider,
    );
    matchQuery.setDefaultDriverAndPassengerParameters({
      driver: defaultParams.DRIVER,
      passenger: defaultParams.PASSENGER,
      seatsProposed: defaultParams.SEATS_PROPOSED,
      seatsRequested: defaultParams.SEATS_REQUESTED,
    });
    expect(matchQuery.seatsProposed).toBe(3);
    expect(matchQuery.seatsRequested).toBe(1);
  });

  it('should set route for a driver only', async () => {
    const matchQuery = new MatchQuery(
      {
        driver: true,
        passenger: false,
        frequency: Frequency.PUNCTUAL,
        fromDate: '2023-08-28',
        toDate: '2023-08-28',
        schedule: [
          {
            time: '01:05',
          },
        ],
        waypoints: [originWaypoint, destinationWaypoint],
      },
      mockRouteProvider,
    );
    await matchQuery.setRoutes();
    expect(matchQuery.driverRoute?.distance).toBe(350101);
    expect(matchQuery.passengerRoute).toBeUndefined();
  });

  it('should set route for a passenger only', async () => {
    const matchQuery = new MatchQuery(
      {
        driver: false,
        passenger: true,
        frequency: Frequency.PUNCTUAL,
        fromDate: '2023-08-28',
        toDate: '2023-08-28',
        schedule: [
          {
            time: '01:05',
          },
        ],
        waypoints: [originWaypoint, destinationWaypoint],
      },
      mockRouteProvider,
    );
    await matchQuery.setRoutes();
    expect(matchQuery.passengerRoute?.distance).toBe(340102);
    expect(matchQuery.driverRoute).toBeUndefined();
  });

  it('should set route for a driver and passenger', async () => {
    const matchQuery = new MatchQuery(
      {
        driver: true,
        passenger: true,
        frequency: Frequency.PUNCTUAL,
        fromDate: '2023-08-28',
        toDate: '2023-08-28',
        schedule: [
          {
            time: '01:05',
          },
        ],
        waypoints: [originWaypoint, destinationWaypoint],
      },
      mockRouteProvider,
    );
    await matchQuery.setRoutes();
    expect(matchQuery.driverRoute?.distance).toBe(350101);
    expect(matchQuery.passengerRoute?.distance).toBe(350101);
  });

  it('should set route for a driver and passenger with 3 waypoints', async () => {
    const matchQuery = new MatchQuery(
      {
        driver: true,
        passenger: true,
        frequency: Frequency.PUNCTUAL,
        fromDate: '2023-08-28',
        toDate: '2023-08-28',
        schedule: [
          {
            time: '01:05',
          },
        ],
        waypoints: [
          originWaypoint,
          intermediateWaypoint,
          { ...destinationWaypoint, position: 2 },
        ],
      },
      mockRouteProvider,
    );
    await matchQuery.setRoutes();
    expect(matchQuery.driverRoute?.distance).toBe(350101);
    expect(matchQuery.passengerRoute?.distance).toBe(340102);
  });

  it('should throw an exception if route is not found', async () => {
    const matchQuery = new MatchQuery(
      {
        driver: true,
        passenger: false,
        frequency: Frequency.PUNCTUAL,
        fromDate: '2023-08-28',
        toDate: '2023-08-28',
        schedule: [
          {
            time: '01:05',
          },
        ],
        waypoints: [originWaypoint, destinationWaypoint],
      },
      mockRouteProvider,
    );
    await expect(matchQuery.setRoutes()).rejects.toBeInstanceOf(Error);
  });
});
